@@ -19,6 +19,7 @@
  */
 
 #include <string>
+#include <sstream>
 #include <mutex>
 #include <iterator>
 #include <algorithm>
@@ -90,6 +91,8 @@ static const std::vector<const char*>& s_mime_image()
 static const char s_mime_gnome_copied_files[] = "x-special/gnome-copied-files";
 static const char s_mime_mate_copied_files[] = "x-special/mate-copied-files";
 
+static const char s_mime_freerdp_update[] = "x-special/freerdp-clipboard-update";
+
 static const char* s_type_HtmlFormat = "HTML Format";
 static const char* s_type_FileGroupDescriptorW = "FileGroupDescriptorW";
 
@@ -127,9 +130,14 @@ static bool operator==(const CLIPRDR_FORMAT& lhs, const CLIPRDR_FORMAT& rhs)
 
 sdlClip::sdlClip(SdlContext* sdl)
     : _sdl(sdl), _file(cliprdr_file_context_new(this)), _log(WLog_Get(TAG)),
-      _system(ClipboardCreate()), _event(CreateEventA(nullptr, TRUE, FALSE, nullptr))
+      _system(ClipboardCreate()), _event(CreateEventA(nullptr, TRUE, FALSE, nullptr)),
+      _uuid(sdl::utils::generate_uuid_v4())
 {
 	WINPR_ASSERT(sdl);
+
+	std::stringstream ss;
+	ss << s_mime_freerdp_update << "-" << _uuid;
+	_mime_uuid = ss.str();
 }
 
 sdlClip::~sdlClip()
@@ -164,12 +172,23 @@ BOOL sdlClip::uninit(CliprdrClientContext* clip)
 	return TRUE;
 }
 
+bool sdlClip::contains(const char** mime_types, Sint32 count)
+{
+	for (Sint32 x = 0; x < count; x++)
+	{
+		const auto mime = mime_types[x];
+		if (mime && (strcmp(_mime_uuid.c_str(), mime) == 0))
+			return true;
+	}
+	return false;
+}
+
 bool sdlClip::handle_update(const SDL_ClipboardEvent& ev)
 {
 	if (!_ctx || !_sync || ev.owner)
 	{
-		/* TODO: Hack to identify our own updates */
-		if (ev.owner && (ev.reserved == 0x42))
+		_last_timestamp = ev.timestamp;
+		if (!_current_mimetypes.empty())
 		{
 			_cache_data.clear();
 			auto rc =
@@ -181,6 +200,16 @@ bool sdlClip::handle_update(const SDL_ClipboardEvent& ev)
 		return true;
 	}
 
+	if (ev.timestamp == _last_timestamp)
+	{
+		return true;
+	}
+
+	if (contains(ev.mime_types, ev.num_mime_types))
+	{
+		return true;
+	}
+
 	clearServerFormats();
 
 	std::string mime_html = s_mime_html;
@@ -198,7 +227,7 @@ bool sdlClip::handle_update(const SDL_ClipboardEvent& ev)
 	size_t nformats = WINPR_ASSERTING_INT_CAST(size_t, ev.num_mime_types);
 	const char** clipboard_mime_formats = ev.mime_types;
 
-	WLog_Print(_log, WLOG_TRACE, "SDL has %d formats", nformats);
+	WLog_Print(_log, WLOG_TRACE, "SDL has %" PRIuz " formats", nformats);
 
 	bool textPushed = false;
 	bool imgPushed = false;
@@ -519,15 +548,15 @@ UINT sdlClip::ReceiveServerFormatList(CliprdrClientContext* context,
 		clipboard->_current_mimetypes.push_back(s_mime_gnome_copied_files);
 		clipboard->_current_mimetypes.push_back(s_mime_mate_copied_files);
 	}
+	clipboard->_current_mimetypes.push_back(clipboard->_mime_uuid.c_str());
 
 	auto s = clipboard->_current_mimetypes.size();
 	SDL_Event ev = { SDL_EVENT_CLIPBOARD_UPDATE };
 	ev.clipboard.owner = true;
+	ev.clipboard.timestamp = SDL_GetTicksNS();
 	ev.clipboard.num_mime_types = WINPR_ASSERTING_INT_CAST(Sint32, s);
 	ev.clipboard.mime_types = clipboard->_current_mimetypes.data();
 
-	/* TODO: Hack to identify our own updates */
-	ev.clipboard.reserved = 0x42;
 	auto rc = (SDL_PushEvent(&ev) == 1);
 	return clipboard->SendFormatListResponse(rc);
 }
