@@ -121,6 +121,9 @@ static int freerdp_connect_begin(freerdp* instance)
 	if (status)
 		status = freerdp_settings_enforce_monitor_exists(settings);
 
+	if (status)
+		status = freerdp_settings_enforce_consistency(settings);
+
 	if (status)
 		status = freerdp_settings_check_client_after_preconnect(settings);
 
@@ -387,16 +390,16 @@ DWORD freerdp_get_event_handles(rdpContext* context, HANDLE* events, DWORD count
 	WINPR_ASSERT(context->rdp);
 	WINPR_ASSERT(events || (count == 0));
 
-	nCount += transport_get_event_handles(context->rdp->transport, events, count);
-
-	if (nCount == 0)
+	const size_t rrc = rdp_get_event_handles(context->rdp, &events[nCount], count - nCount);
+	if (rrc == 0)
 		return 0;
 
+	nCount += WINPR_ASSERTING_INT_CAST(uint32_t, rrc);
+
 	if (events && (nCount < count + 2))
 	{
 		events[nCount++] = freerdp_channels_get_event_handle(context->instance);
 		events[nCount++] = getChannelErrorEventHandle(context);
-		events[nCount++] = utils_get_abort_event(context->rdp);
 	}
 	else
 		return 0;
@@ -786,6 +789,56 @@ BOOL freerdp_context_new(freerdp* instance)
 	return freerdp_context_new_ex(instance, NULL);
 }
 
+static BOOL freerdp_common_context(rdpContext* context, AccessTokenType tokenType, char** token,
+                                   size_t count, ...)
+{
+	BOOL rc = FALSE;
+
+	WINPR_ASSERT(context);
+	if (!context->instance || !context->instance->GetAccessToken)
+		return TRUE;
+
+	va_list ap;
+	va_start(ap, count);
+	switch (tokenType)
+	{
+		case ACCESS_TOKEN_TYPE_AAD:
+			if (count != 2)
+			{
+				WLog_ERR(TAG,
+				         "ACCESS_TOKEN_TYPE_AAD expected 2 additional arguments, but got %" PRIuz
+				         ", aborting",
+				         count);
+			}
+			else
+			{
+				const char* scope = va_arg(ap, const char*);
+				const char* req_cnf = va_arg(ap, const char*);
+				rc = context->instance->GetAccessToken(context->instance, tokenType, token, count,
+				                                       scope, req_cnf);
+			}
+			break;
+		case ACCESS_TOKEN_TYPE_AVD:
+			if (count != 0)
+			{
+				WLog_WARN(TAG,
+				          "ACCESS_TOKEN_TYPE_AVD expected 0 additional arguments, but got %" PRIuz
+				          ", ignoring",
+				          count);
+			}
+			else
+			{
+				rc = context->instance->GetAccessToken(context->instance, tokenType, token, count);
+			}
+			break;
+		default:
+			break;
+	}
+	va_end(ap);
+
+	return rc;
+}
+
 BOOL freerdp_context_new_ex(freerdp* instance, rdpSettings* settings)
 {
 	rdpRdp* rdp = NULL;
@@ -866,10 +919,19 @@ BOOL freerdp_context_new_ex(freerdp* instance, rdpSettings* settings)
 	if (!context->dump)
 		goto fail;
 
+	/* Fallback:
+	 * Client common library might set a function pointer to handle this, but here we provide a
+	 * default implementation that simply calls instance->GetAccessToken.
+	 */
+	if (!freerdp_set_common_access_token(context, freerdp_common_context))
+		goto fail;
+
 	IFCALLRET(instance->ContextNew, ret, instance, context);
 
-	if (ret)
-		return TRUE;
+	if (!ret)
+		goto fail;
+
+	return TRUE;
 
 fail:
 	freerdp_context_free(instance);
@@ -1080,26 +1142,26 @@ void freerdp_set_last_error_ex(rdpContext* context, UINT32 lastError, const char
 	{
 		if (WLog_IsLevelActive(context->log, WLOG_ERROR))
 		{
-			WLog_PrintMessage(context->log, WLOG_MESSAGE_TEXT, WLOG_ERROR, (size_t)line, file, fkt,
-			                  "%s [0x%08" PRIX32 "]", freerdp_get_last_error_name(lastError),
-			                  lastError);
+			WLog_PrintTextMessage(context->log, WLOG_ERROR, (size_t)line, file, fkt,
+			                      "%s [0x%08" PRIX32 "]", freerdp_get_last_error_name(lastError),
+			                      lastError);
 		}
 	}
 
 	if (lastError == FREERDP_ERROR_SUCCESS)
 	{
 		if (WLog_IsLevelActive(context->log, WLOG_DEBUG))
-			WLog_PrintMessage(context->log, WLOG_MESSAGE_TEXT, WLOG_DEBUG, (size_t)line, file, fkt,
-			                  "resetting error state");
+			WLog_PrintTextMessage(context->log, WLOG_DEBUG, (size_t)line, file, fkt,
+			                      "resetting error state");
 	}
 	else if (context->LastError != FREERDP_ERROR_SUCCESS)
 	{
 		if (WLog_IsLevelActive(context->log, WLOG_ERROR))
 		{
-			WLog_PrintMessage(context->log, WLOG_MESSAGE_TEXT, WLOG_ERROR, (size_t)line, file, fkt,
-			                  "TODO: Trying to set error code %s, but %s already set!",
-			                  freerdp_get_last_error_name(lastError),
-			                  freerdp_get_last_error_name(context->LastError));
+			WLog_PrintTextMessage(context->log, WLOG_ERROR, (size_t)line, file, fkt,
+			                      "TODO: Trying to set error code %s, but %s already set!",
+			                      freerdp_get_last_error_name(lastError),
+			                      freerdp_get_last_error_name(context->LastError));
 		}
 	}
 	context->LastError = lastError;
@@ -1504,3 +1566,19 @@ const char* freerdp_disconnect_reason_string(int reason)
 			return "rn-unknown";
 	}
 }
+
+BOOL freerdp_set_common_access_token(rdpContext* context,
+                                     pGetCommonAccessToken GetCommonAccessToken)
+{
+	WINPR_ASSERT(context);
+	WINPR_ASSERT(context->rdp);
+	context->rdp->GetCommonAccessToken = GetCommonAccessToken;
+	return TRUE;
+}
+
+pGetCommonAccessToken freerdp_get_common_access_token(rdpContext* context)
+{
+	WINPR_ASSERT(context);
+	WINPR_ASSERT(context->rdp);
+	return context->rdp->GetCommonAccessToken;
+}
